from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QDialog
from PyQt5.QtGui import QColor
from PyQt5.QtWidgets import QApplication, QWidget, QGridLayout, QPushButton
from ...plotters.widgets.signal_line import SignalLine


N_ROWS = 3


class SignalUnitSelector(QDialog):
    NEW = 9999
    STIMULUS = 1000
    NOISE = -1
    CANCEL = -9999

    def __init__(self, signal_ids, parent=None):
        super().__init__(parent, Qt.CustomizeWindowHint)

        app = QApplication.instance()
        main_app_stylesheet = app.activeWindow().styleSheet()

        self.setStyleSheet(main_app_stylesheet)
        self.setAttribute(Qt.WA_TranslucentBackground)

        # Set the window title and message text
        self.setWindowTitle("Manual sorting")

        widget = QWidget()
        layout = QGridLayout(widget)
        self.setLayout(layout)

        # Add buttons to the grid
        signal_ids = sorted([sid for sid in signal_ids if (sid >= 0) and (sid < SignalUnitSelector.STIMULUS)])

        for ix, signal_id in enumerate(signal_ids):
            btn = QPushButton(f"{signal_id}")
            self._style_button(btn, QColor(*SignalLine.color(signal_id)))

            layout.addWidget(btn, int(ix/N_ROWS), ix % N_ROWS)
            btn.clicked.connect(lambda _, btn=btn: self.handle_button_click(btn))

        btn_noise = QPushButton(f"Noise")
        self._style_button(btn_noise, QColor(*SignalLine.color(-1)))

        btn_stim = QPushButton(f"Stim")
        self._style_button(btn_stim, QColor(*SignalLine.color(SignalUnitSelector.STIMULUS)))

        btn_new = QPushButton(f"New")
        self._style_button(btn_new, QColor(255, 255, 255))

        layout.addWidget(btn_noise, int(ix/N_ROWS) + 1, 0)
        layout.addWidget(btn_stim, int(ix/N_ROWS) + 1, 1)
        layout.addWidget(btn_new, int(ix/N_ROWS) + 1, 2)

        btn_noise.clicked.connect(lambda _, btn=btn_noise: self.handle_button_click(btn))
        btn_stim.clicked.connect(lambda _, btn=btn_stim: self.handle_button_click(btn))
        btn_new.clicked.connect(lambda _, btn=btn_new: self.handle_button_click(btn))

        btn_cancel = QPushButton("Cancel", self)
        self._style_button(btn_cancel, QColor(255, 255, 255))

        layout.addWidget(btn_cancel, int(ix/N_ROWS) + N_ROWS, 0, 1, N_ROWS)  # Span two columns for the Cancel button

        # Connect the Cancel button click to the reject slot
        btn_cancel.clicked.connect(self.reject)
        btn_cancel.clicked.connect(lambda _, btn=btn_cancel: self.handle_button_click(btn))

        # Set the widget as the message box's layout
        self.layout().addWidget(widget)

    def _style_button(self, btn, background_color):
        style = [f"background-color: {background_color.name()}",
                 "width: 50px",
                 f"color: {background_color.darker(240).name()}",
                 f"border: 2px solid {background_color.darker(120).name()}"]
        btn.setStyleSheet(";".join(style))

    def handle_button_click(self, btn):
        if btn.text().upper() == 'NOISE':
            self.done(SignalUnitSelector.NOISE)
        elif btn.text().upper() == 'NEW':
            self.done(SignalUnitSelector.NEW)
        elif btn.text().upper() == 'CANCEL':
            self.done(SignalUnitSelector.CANCEL)
        elif btn.text().upper() == 'STIM':
            self.done(SignalUnitSelector.STIMULUS)
        else:
            self.done(int(float(btn.text())))



import numpy as np
from PyQt5.QtChart import QChart, QChartView, QValueAxis
from PyQt5.QtCore import Qt, QPointF, pyqtSignal
from PyQt5.QtGui import QPainter
from PyQt5.QtWidgets import QWidget, QVBoxLayout


from ..plotters.widgets.signal_line import SignalLine
from ..plotters.widgets.signal_unit_selector import SignalUnitSelector
from ..plotters.widgets.signal_view_selector import SignalViewSelector

from ..utils.logger import Logger
logger = Logger.get_logger()


class DragSelectChartView(QChartView):
    def __init__(self, chart, parent=None):
        super().__init__(chart, parent)

        self.parent = parent

        # Variables to track selection
        self.dragging = False
        self.startPos = None
        self.endPos = None
        self.allow_dragging = True


    def allow_selection(self, allow):
        self.allow_dragging = allow

    def mousePressEvent(self, event):
        if event.button() == Qt.LeftButton:
            self.dragging = True
            self.startPos = event.pos()
            self.endPos = event.pos()
        elif event.button() == Qt.RightButton:
            self.parent.view_series()

    def mouseReleaseEvent(self, event):
        if self.allow_dragging and self.dragging:
            self.endPos = event.pos()
            self.dragging = False

            x_start = self.chart().mapToValue(self.mapToScene(self.startPos)).x()
            y_start = self.chart().mapToValue(self.mapToScene(self.startPos)).y()

            x_end = self.chart().mapToValue(self.mapToScene(self.endPos)).x()
            y_end = self.chart().mapToValue(self.mapToScene(self.endPos)).y()

            x_start, x_end = min(x_start, x_end), max(x_start, x_end)
            y_start, y_end = min(y_start, y_end), max(y_start, y_end)

            self.parent.select_series(x_start, x_end, y_start, y_end)


class Line2D(QWidget):
    signal_unit_changed = pyqtSignal(list)

    def __init__(self, parent):
        super().__init__(parent)

        self.chart = QChart()
        self.series = []

        self.xaxis = QValueAxis()
        self.xaxis.setTickCount(11)
        self.yaxis = QValueAxis()
        self.yaxis.setTickCount(11)


        self.chart.addAxis(self.xaxis, Qt.AlignBottom)
        self.chart.addAxis(self.yaxis, Qt.AlignLeft)
        self.chart.legend().setVisible(False)  # Hide the plot legend

        self.single_chart_view = DragSelectChartView(self.chart, parent=self)
        self.single_chart_view.setRenderHint(QPainter.Antialiasing)  # Optional: enable antialiasing for smoother lines

        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)  # Set margins to zero
        layout.setSpacing(0)
        layout.addWidget(self.single_chart_view)

        parent.setLayout(layout)

    def x_label(self, xlab):
        self.xaxis.setLabelFormat("%.2f")  # Set the label format for the X axis
        self.xaxis.setTitleText(xlab)  # Set the label for the X axis
        return self

    def y_label(self, ylab):
        self.yaxis.setLabelFormat("%.2f")
        self.yaxis.setTitleText(ylab)
        return self

    def title(self, title):
        self.chart.setTitle(title)
        return self

    def plot_xy(self, x_values, y_values, *args, **kwargs):
        self.plot([zip(x_values, y_values)], *args, **kwargs)

    def plot(self,  values=None, group=None, allow_selection=False, xlim=None, ylim=None, progress_bar=False):
        self.single_chart_view.allow_selection(allow_selection)
        self.clear()

        y_range = (np.Inf, -np.Inf)
        x_range = (np.Inf, -np.Inf)
        for ix, series in enumerate(values):
            if progress_bar:
                logger.log_process('plot', ix, len(values))

            line = SignalLine()
            if group is not None:
                line.set_id(group[ix])
            line.append([QPointF(x, y) for x, y in series])
            line.set_index(ix)

            x_range, y_range = self._update_ranges(line, x_range, y_range)

            self.chart.addSeries(line)
            self.chart.setAxisY(self.yaxis, line)
            self.chart.setAxisX(self.xaxis, line)

            self.series.append(line)

        if progress_bar:
            logger.log_process('plot', len(values), len(values))

        x_range = x_range if xlim is None else xlim
        y_range = y_range if ylim is None else ylim
        self.xaxis.setRange(*x_range)
        self.yaxis.setRange(*y_range)

    @classmethod
    def _update_ranges(cls, line, x_range, y_range):
        y_r = line.y_range
        y_range = (min(y_r[0], y_range[0]), max(y_r[1], y_range[1]))
        x_r = line.x_range
        x_range = (min(x_r[0], x_range[0]), max(x_r[1], x_range[1]))
        return x_range, y_range

    def clear(self):
        for series in self.series:
            self.chart.removeSeries(series)
            series.clear()
        self.series.clear()
        self.series = []

    def select_series(self, x_start, x_end, y_start, y_end):
        selected_series = []
        signal_ids = set()

        for series in self.series:
            signal_ids.add(series.signal_id)
            if series.has_points_inside_bounding_box(x_start, x_end, y_start, y_end):
                selected_series.append(series)
                series.select()

        if selected_series:
            msg_box = SignalUnitSelector(signal_ids)
            result = msg_box.exec()
            time_unit = []
            if result == SignalUnitSelector.CANCEL:
                for series in selected_series:
                    series.unselect()
            elif result == SignalUnitSelector.NEW:
                new_id = sorted(list(set(range(1, 100)).difference(set(signal_ids))))[0]
                for series in selected_series:
                    series.set_id(new_id)
                    time_unit.append((series.index, series.signal_id))
            else:
                for series in selected_series:
                    series.set_id(result)
                    time_unit.append((series.index, series.signal_id))

            self.signal_unit_changed.emit(time_unit)

        self.update()

    def view_series(self):

        signal_ids = set()

        for series in self.series:
            signal_ids.add(series.signal_id)
        msg_box = SignalViewSelector(signal_ids)
        result = msg_box.exec()
        if result == 1:
            signal_ids_selected = msg_box.get_selected()

            y_min, y_max = np.Inf, -np.Inf
            for series in self.series:
                visible = signal_ids_selected.get(series.signal_id, True)
                series.set_visible(visible)
                if visible:
                    y_range = series.y_range
                    y_min = min(y_min, y_range[0])
                    y_max = max(y_max, y_range[1])

            self.yaxis.setRange(y_min, y_max)
            self.update()



# logger
from ..utils.logger import Logger

# pyqt
from PyQt5 import QtWidgets, uic
from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QMessageBox

from ..utils.singleton import Singleton
from ..config.layout import Layout
from ..recordings.config import UI_LAYOUT_CREATION


class LayoutCreationUi(QtWidgets.QDialog):
    def __init__(self):
        super(LayoutCreationUi, self).__init__()
        self.setWindowFlags(Qt.CustomizeWindowHint | Qt.WindowCloseButtonHint)

        uic.loadUi(UI_LAYOUT_CREATION, self)

        self.show()

    def create_layout(self):
        self.done(1)

    def get_data(self):
        name = self.layout_name.text()
        cols, rows = self.num_cols.value(), self.num_rows.value()
        return name, cols, rows


class LayoutUi(metaclass=Singleton):

    def __init__(self, ui):
        self.ui = ui
        self.layout = Layout()

    def new_layout(self):
        layout_ui = LayoutCreationUi()
        result = layout_ui.exec_()

        if result == QtWidgets.QDialog.Accepted:
            name, cols, rows = layout_ui.get_data()
            self.layout.new(name, cols, rows)
            self.load_layout_names(name)

    def delete_layout(self):
        layout_name = self.ui.layoutSelector.currentText()
        if layout_name:
            result = QMessageBox.question(self.ui, "Confirmation", "Are you sure you want to delete this item?",
                                          QMessageBox.Yes | QMessageBox.No, QMessageBox.No)

            if result == QMessageBox.Yes:
                self.layout.delete(layout_name)
                self.load_layout_names()

    def set_current_layout(self, selected):
        if selected > 0:
            layout_name = self.ui.layoutSelector.currentText()
            self.layout.set_current_layout(layout_name)

    def select_layout(self, selected_name):
        layout_names = self.layout.names()

        selected_ix = 0
        for ix, name in enumerate(layout_names):
            if name == selected_name:
                selected_ix = ix
        if len(layout_names) > 0:
            self.ui.layoutSelector.setCurrentIndex(selected_ix)
            self.plot_layout(layout_names[selected_ix])
            self.ui.currentLayoutCheck.setChecked(self.layout.is_current_layout(layout_names[selected_ix]))

    def load_layout_names(self, selected_name=None):
        layout_names = self.layout.names()
        self.ui.layoutSelector.clear()
        for ix, name in enumerate(layout_names):
            self.ui.layoutSelector.addItem(name)
        self.select_layout(selected_name)

    def plot_layout(self, name):
        self.ui.layoutTbl.itemChanged.disconnect()

        cols, rows = self.layout.get_shape(name)
        cell_values = self.layout.get_layout(name)
        column_headers = self.layout.column_headers(cols)
        row_headers = self.layout.row_headers(rows)

        self.ui.layoutTbl.setRowCount(rows)
        self.ui.layoutTbl.setColumnCount(cols)

        for col, label in enumerate(column_headers):
            self.ui.layoutTbl.setHorizontalHeaderItem(col, QtWidgets.QTableWidgetItem(label))
        for row, label in enumerate(row_headers):
            self.ui.layoutTbl.setVerticalHeaderItem(row, QtWidgets.QTableWidgetItem(label))

        for col in range(cols):
            self.ui.layoutTbl.setColumnWidth(col, 80)
            for row in range(rows):
                item = QtWidgets.QTableWidgetItem(cell_values[row][col])
                item.setFlags(item.flags() | Qt.ItemIsEditable)  # Make the item editable
                self.ui.layoutTbl.setItem(row, col, item)

        self.ui.layoutTbl.itemChanged.connect(self.handle_table_item_change)

    def handle_table_item_change(self, item):
        layout_name = self.ui.layoutSelector.currentText()
        tab = self.ui.layoutTbl
        col, row = tab.columnCount(), tab.rowCount()

        cell_values = [[tab.item(r, c).text() for c in range(col)] for r in range(row)]

        self.layout.update(layout_name, cell_values)
        self.plot_layout(layout_name)


from PyQt5 import QtWidgets
import glob

import numpy as np

from ..recordings.mcs_raw import McsRaw
from ..ui.abstract_source import AbstractSource
from ..config.settings import Settings
settings = Settings()

from ..utils.logger import Logger


class Raw(AbstractSource):
    """
    A class for handling raw data visualization and spike extraction.

    This class handles the visualization and processing of raw data, including
    loading files, plotting different views and extracting spikes.

    Methods:
        __init__(self, ui): Initializes the Raw instance.
        load_files(self, folder_path): Loads file names from a folder.
        select_view(self): Selects the raw data view.
        _plot_single_view(self, path, channel_id): Plots a single channel view.
        _plot_grid_view(self, path): Plots a grid view of multiple channels.
        _plot_list_view(self, channel_id): Plots a list view of channels.
        get_filename(cls, path): Extracts the filename from a path.
        extract_spikes(self): Extracts spikes from the selected file.
        extract_all_spikes(self): Extracts spikes from all files.

    """
    def __init__(self, ui):
        super().__init__(ui, McsRaw)

    def load_files(self, folder_path):
        """
        Load spike files from the given folder path.

        Args:
            folder_path (str): Path to the folder containing h5 files.
        """
        self.ui.rawBtn.setEnabled(False)

        self.filenames = sorted(glob.glob(f"{folder_path}/*.h5"))

        fns = [self.get_filename(path) for path in self.filenames]
        self.ui.list_chart_view.setup_grid(nrows=int(np.ceil(len(self.filenames)/2)), ncols=2, plot_names=fns
                                           , show_axis=True)

        if len(self.filenames) > 0:
            self.ui.rawBtn.setEnabled(True)

    def select_view(self):
        """
        Selects and displays the raw view.

        This method is called to select the raw view and update the user interface
        accordingly. It sets the current widget of the option stack to the raw
        options page.

        Usage:
            Call this method when you want to switch to the raw view.
        """
        super().select_view()
        self.ui.optionStck.setCurrentWidget(self.ui.rawOptPage)

    def _plot_single_view(self, path, channel_id):
        channel_id = int(channel_id)
        mcs = self.data_handler(path)

        start_idx = mcs.to_ticks(self.ui.recording_start_s, 'second')
        end_idx = mcs.to_ticks(self.ui.recording_start_s + settings.ui_raw.SINGLE_VIEW_LENGTH_S, 'second')
        time, series = mcs.get_channel_data(channel_id, start_idx, end_idx)
        (
            self.ui.single_chart_view
            .x_label('Time (seconds)')
            .y_label('Microvolts')
            .title(f'Channel {channel_id}')
            .plot_xy(time, series, ylim=settings.ui_raw.MICRO_VOLTS_RANGE)
        )

    def _plot_grid_view(self, path):
        mcs = self.data_handler(path)
        start_idx = mcs.to_ticks(self.ui.recording_start_s, 'second')
        end_idx = mcs.to_ticks(self.ui.recording_start_s + settings.ui_raw.GRID_VIEW_LENGTH_S, 'second')
        channel_data = mcs.get_all_data(start_idx, end_idx)
        (
            self.ui.multiple_chart_view
            .plot_xy(channel_data, ylim=settings.ui_raw.MICRO_VOLTS_RANGE)
        )

    def _plot_list_view(self, channel_id):
        data = dict()
        for path in self.filenames:
            mcs = self.data_handler(path)
            start_idx = mcs.to_ticks(self.ui.recording_start_s, 'second')
            end_idx = mcs.to_ticks(self.ui.recording_start_s + settings.ui_raw.GRID_VIEW_LENGTH_S, 'second')
            channel_data = mcs.get_channel_data(channel_id, start_idx, end_idx)
            data[self.get_filename(path)] = channel_data
        (
            self.ui.list_chart_view
            .plot_xy(data)
        )

    @classmethod
    def get_filename(cls, path):
        """
        Extracts the filename from the given path.

        Args:
            path: The full path to the file.

        Returns:
            str: The extracted filename.
        """
        return path.split('/')[-1].replace('.h5', '')

    def extract_spikes(self):
        """
        Extracts spike events from the selected data file and saves them to a spike file.

        This method extracts spike events from the currently selected data file, using the specified parameters for
        spike detection.

        The spike events are detected based on the given sigma and window values. A Gaussian filter with the specified
        sigma is applied to the raw data, and threshold crossings exceeding the specified window duration are
        considered as spike events.

        After extracting spike events, the resulting spike events are saved to a CSV file in the same directory as the
        data file. The filename for the spike file is derived from the original data file name by replacing the '.h5'
        extension with '.spike'.

        Note:
            - The data file must be selected in the UI's file selector before calling this method.
            - The sigma and window parameters are determined by the UI's corresponding user inputs.

        """
        path = self.ui.fileSelector.currentText()
        sigma = self.ui.sigmaValue.value()
        window_ms = self.ui.windowValue.value()
        if path:
            mcs = self.data_handler(path)
            spikes = mcs.extract_all_spikes(sigma, window_ms)
            spikes.to_csv(path.replace('h5', 'spike'), index=False)

    def extract_all_spikes(self):
        """
        Extracts spike events from all data files and saves them to individual spike files.

        This method iterates through all available data files in the file selector of the UI and extracts spike events
        from each file. Spike detection is performed using the specified parameters for spike detection, such as sigma
        and window duration.

        For each data file, spike events are detected based on the given parameters.

        After extracting spike events for each data file, the resulting spike events are saved to individual CSV files
        in the same directories as the corresponding data files. The filenames for the spike files are derived from the
        original data file names by replacing the '.h5' extension with '.spike'.

        Note:
            - The sigma and window parameters are determined by the UI's corresponding user inputs.

        """
        num_items = self.ui.fileSelector.count()
        for i in range(num_items):
            self.ui.fileSelector.setCurrentIndex(i)
            self.ui.update()
            QtWidgets.QApplication.processEvents()
            self.extract_spikes()




import unittest
from unittest.mock import patch, MagicMock
from neuronautics.analysis.type.graph_analysis import GraphAnalysis
from neuronautics.analysis.type.abstract_analysis import AnalysisType
from matplotlib.figure import Figure
import pandas as pd
import numpy as np


class OneGraphAnalysis(GraphAnalysis):

    def run(self, *args, **kwargs):
        return np.array([[0, 1], [1, 0]])

    def get_input_params(self):
        pass


class TestGraphAnalysis(unittest.TestCase):
    def test_type(self):
        analysis = OneGraphAnalysis()
        self.assertEqual(analysis.type(), AnalysisType.GRAPH)

    @patch("networkx.draw")
    def test_plot(self, mock_draw):
        # Act
        analysis = OneGraphAnalysis()
        result = analysis.plot(title='my plot')

        # Assert
        self.assertIsInstance(result, Figure)
        mock_draw.assert_called_once()


if __name__ == '__main__':
    unittest.main()



import unittest
from unittest.mock import patch, MagicMock
import numpy as np
from numpy import testing
from neuronautics.utils.helpers import (moving_stats, parse_numeric_array,
                                        mkdir, app_path, file_path, load_yaml,
                                        open_external_editor)


class HelpersTest(unittest.TestCase):
    def test_moving_stats(self):
        data = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])

        expected_avg = np.array([2., 3., 4., 5., 6., 7., 8., 9.])
        expected_std = np.array([0.816496580927726]*8)

        avg, std = moving_stats(data, window=3)

        testing.assert_array_equal(avg, expected_avg)
        testing.assert_array_equal(std, expected_std)

    def test_moving_stats_insufficient_data(self):
        data = np.array([])

        with self.assertRaises(ValueError) as context:
            _ = moving_stats(data, window=3)
            self.assertTrue('array must be longer' in context.exception)

    def test_moving_stats_empty_data(self):
        data = None

        with self.assertRaises(ValueError) as context:
            _ = moving_stats(data, window=3)
            self.assertTrue('array must be longer' in context.exception)

    def test_moving_stats_small_window(self):
        data = np.array([1, 2, 3])

        with self.assertRaises(ValueError) as context:
            _ = moving_stats(data, window=2)
            self.assertTrue('window must be greater' in context.exception)

    def test_parse_numeric_array(self):
        input_str = "[  0. -0.29   0.72,   -100.  100. ]"
        expected_array = [0., -0.29, 0.72, -100.,  100.]

        array = parse_numeric_array(input_str)

        testing.assert_array_equal(array, expected_array)

    def test_parse_numeric_array_empty_array(self):
        input_str = "[  ]"
        expected_array = []

        array = parse_numeric_array(input_str)

        testing.assert_array_equal(array, expected_array)

    def test_parse_numeric_array_no_brackets(self):
        input_str = "0. -0.29   0.72   -100.  100."
        expected_array = [0., -0.29, 0.72, -100.,  100.]

        array = parse_numeric_array(input_str)

        testing.assert_array_equal(array, expected_array)

    def test_parse_numeric_array_error(self):
        input_str = [1,2,3]
        expected_array = []

        array = parse_numeric_array(input_str)

        testing.assert_array_equal(array, expected_array)

    def test_mkdir(self):
        with patch('neuronautics.utils.helpers.Path') as spy_path:
            mock_path_instance = MagicMock()
            spy_path.return_value = mock_path_instance

            mock_path_instance.mkdir.return_value = None
            new_directory = mkdir('path', 'to', 'my', 'test_directory')

            self.assertEqual(new_directory, 'path/to/my/test_directory')
            spy_path.assert_called_once_with('path/to/my/test_directory')
            mock_path_instance.mkdir.assert_called_once_with(parents=True, exist_ok=True)

    def test_app_path(self):
        with (patch('neuronautics.utils.helpers.mkdir') as spy_mkdir,
              patch('neuronautics.utils.helpers.os.path.expanduser') as spy_os_path):
            spy_os_path.return_value = '/home/fakeuser'
            spy_mkdir.return_value = 'mkdir output'

            directory = app_path('folder1')

            spy_mkdir.assert_called_once_with('/home/fakeuser', '.neuronautics', 'folder1')
            self.assertEqual(directory, 'mkdir output')

    def test_file_path(self):
        with patch('neuronautics.utils.helpers.app_path') as spy_app_path:
            spy_app_path.return_value = '/app/path/extra/path/to'

            filepath = file_path('extra/path/to/file.txt')

            spy_app_path.assert_called_once_with('extra', 'path', 'to')
            self.assertEqual(filepath, '/app/path/extra/path/to/file.txt')

    def test_load_yaml_file_not_found(self):
        with patch('neuronautics.utils.helpers.Path') as spy_path:
            mock_path_instance = MagicMock()
            spy_path.return_value = mock_path_instance
            mock_path_instance.exists.return_value = False

            content = load_yaml('random.yml', 'expected output')

            self.assertEqual(content, 'expected output')

    def test_load_yaml_file_found(self):
        with (patch('neuronautics.utils.helpers.Path') as spy_path,
              patch('builtins.open') as spy_open,
              patch('yaml.full_load') as spy_yaml):
            mock_path_instance = MagicMock()
            spy_path.return_value = mock_path_instance
            mock_path_instance.exists.return_value = True

            spy_open = MagicMock()
            spy_open.return_value = 'stream'
            spy_yaml.return_value = {'random': 'dict'}

            content = load_yaml('random.yml', 'unexpected output')

            spy_path.assert_called_once_with('random.yml')

            self.assertDictEqual(content, {'random': 'dict'})

    def test_load_yaml_file_empty_data(self):
        with (patch('neuronautics.utils.helpers.Path') as spy_path,
              patch('builtins.open') as spy_open,
              patch('yaml.full_load') as spy_yaml):
            mock_path_instance = MagicMock()
            spy_path.return_value = mock_path_instance
            mock_path_instance.exists.return_value = True

            spy_open = MagicMock()
            spy_open.return_value = 'stream'
            spy_yaml.return_value = None

            content = load_yaml('random.yml', 'expected output')

            spy_path.assert_called_once_with('random.yml')

            self.assertEqual(content, 'expected output')

    def test_open_external_editor_existing_os(self):
        with (patch('platform.system') as spy_system,
              patch('subprocess.run') as spy_run):
            spy_system.return_value = 'Windows'

            open_external_editor('random.txt')

            spy_run.assert_called_with(['code', 'random.txt'], check=True)

    def test_open_external_editor_unknown_os(self):
        with (patch('platform.system') as spy_system,
              patch('subprocess.run') as spy_run,
              patch('builtins.print') as spy_print):
            spy_system.return_value = 'Android'

            open_external_editor('random.txt')

            spy_run.assert_not_called()
            spy_print.assert_called_once_with("Unsupported operating system.")

    def test_open_external_editor_unable_to_open(self):
        import subprocess
        with (patch('platform.system') as spy_system,
              patch('subprocess.run') as spy_run,
              patch('builtins.print') as spy_print):
            spy_system.return_value = 'Linux'
            spy_run.side_effect = subprocess.CalledProcessError(returncode=1, cmd='mock command')

            open_external_editor('random.txt')

            spy_run.assert_any_call(['xdg-open', 'random.txt'], check=True)
            spy_print.assert_called_once_with("Unable to open the file in any editor.")




if __name__ == '__main__':
    unittest.main()



import unittest
from neuronautics.utils.singleton import Singleton


class TestSingleton(unittest.TestCase):
    def test_singleton(self):
        class A(metaclass=Singleton):
            pass

        instance1 = A()
        instance2 = A()
        self.assertEqual(instance1, instance2)


if __name__ == '__main__':
    unittest.main()



import unittest
from unittest.mock import  MagicMock
from neuronautics.ui.helpers import clear_layout


class TestHelpers(unittest.TestCase):

    def test_clear_layout(self):
        # Create a mock QLayout and its associated widget
        layout_mock = MagicMock()
        widget_mock = MagicMock()
        layout_mock.__mock=[MagicMock(widget=widget_mock), MagicMock(widget=widget_mock)]
        layout_mock.count.side_effect = lambda: len(layout_mock.__mock)
        layout_mock.takeAt.side_effect = lambda i: layout_mock.__mock.pop(i)

        # Call the function with the mock layout
        clear_layout(layout_mock)

        # Verify that methods were called as expected
        layout_mock.takeAt.assert_called_with(0)
        for w in layout_mock.__mock:
            w.deleteLater.assert_called()


if __name__ == '__main__':
    unittest.main()
